import yaml
import argparse
import re
import json
from pathlib import Path

parser = argparse.ArgumentParser(description="Process some integers.")
parser.add_argument("file", type=Path, help="yaml config")
parser.add_argument("--json", action="store_true", help="output parse as json")
parser.add_argument("--out", type=Path, help="write into file")

args = parser.parse_args()


# ╔───────────────────────────────────────────────────────────────────────────╗
# │ Λssεrτiδηs                                                                |
# ╚───────────────────────────────────────────────────────────────────────────╝
def check_for_duplicates(host, props):
    seen = set()

    for v in props:
        if v in seen:
            raise SyntaxError(f"Duplicate key on `{host}`: `{v}`")
        else:
            seen.add(v)


def check_duplicate_hosts(data):
    hosts = set()

    for host in data:
        if host in hosts:
            raise SyntaxError(f"Duplicate host: `{host}`")
        hosts.add(host)


def run_assertions(host, props):
    check_for_duplicates(host, props)


# ╔───────────────────────────────────────────────────────────────────────────╗
# │ Pαrsεr                                                                    |
# ╚───────────────────────────────────────────────────────────────────────────╝
def expand_for_loop(props):
    expanded_props = dict()

    for loop in props.get("for", []):
        prop = loop["property"]

        if expanded_props.get(prop) is None:
            expanded_props[prop] = list()

        if loop.get("range"):
            r = loop["range"]
            start = int(r[0])
            cease = int(r[1])
            incre = int(r[2]) if len(r) > 2 else 1

            iterations = range(start, cease, incre)
        elif loop.get("iter"):
            iterations = loop["iter"]
        else:
            raise TypeError("Expected a `range` or `iter` property on `for`")

        pattern = re.compile(rf'\$\{{{loop["variable"]}\}}')

        for i in iterations:
            expanded_props[prop].append(pattern.sub(str(i), loop["template"]))

    return expanded_props


def parse_props(host, props, hosts):
    parsed = dict()

    # Standard ssh properties
    for k, v in props.get("ssh_props", {}).items():
        parsed[k] = v if isinstance(v, list) else [v]

    # For-loops in config
    for prop, values in expand_for_loop(props).items():
        if parsed.get(prop) is not None:
            parsed[prop].extend(values)
        else:
            parsed[prop] = values

    # Inherit from template
    template = props.get("template")

    if template and template not in hosts:
        raise SyntaxError(
            f"Template `{template}` required by `{host}` is not declared above `{host}` in the yaml"
        )
    elif template:
        for k, v in hosts[props.get("template")].items():
            if k not in parsed:
                parsed[k] = v

    return parsed


def remove_templates(expanded):
    hosts = dict(expanded)

    for host, props in data.items():
        if host.startswith("__"):
            del hosts[host]

    return hosts


def parse_yaml(data):
    hosts = dict()
    check_duplicate_hosts(data)

    for host, props in data.items():
        hosts[host] = parse_props(host, props, hosts)

    hosts = remove_templates(hosts)
    return hosts


# ╔───────────────────────────────────────────────────────────────────────────╗
# │ Fδrmαττεr                                                                 |
# ╚───────────────────────────────────────────────────────────────────────────╝
def to_ssh_config_string(parse):
    s = str()

    for i, (host, keys) in enumerate(parse.items()):
        lines = list()

        for key, values in keys.items():
            for value in values:
                if isinstance(value, bool):
                    lines.append(f"\t{key} {'true' if value else 'false'}")
                else:
                    lines.append(f"\t{key} {value}")

        # join the sorted lines on a new line character
        s += f"Host {host}\n" + "\n".join(sorted(lines))

        s += "\n\n" if i < len(parse) - 1 else "\n"

    return s


# ╔───────────────────────────────────────────────────────────────────────────╗
# │ Mαiη                                                                      |
# ╚───────────────────────────────────────────────────────────────────────────╝
if __name__ == '__main__':
    with open(args.file, "r") as f:
        data = yaml.safe_load(f)

    parse = parse_yaml(data)

    if args.json:
        print(json.dumps(parse))
    elif args.out:
        with open(args.out, "w") as f:
            f.write(to_ssh_config_string(parse))
    else:
        print(to_ssh_config_string(parse), end="")
